@@ -34,7 +34,7 @@ except ImportError:
     asyncio = None
 
 
-__version__ = '3.2'
+__version__ = '3.4'
 
 log = logging.getLogger(__name__)
 
@@ -58,7 +58,7 @@ IDENTITY_ENCODING = 'identity_encoding'
 WAITING_FOR_100c = object()
 
 #: Sequence of ``(hostname, port)`` tuples that are used by
-#: `is_temp_network_error` to distinguish between permanent and temporary name
+#: dugong to distinguish between permanent and temporary name
 #: resolution problems.
 DNS_TEST_HOSTNAMES=(('www.google.com', 80),
                     ('www.iana.org', 80),
@@ -185,6 +185,36 @@ class _GeneralError(Exception):
     def __str__(self):
         return self.msg
 
+class HostnameNotResolvable(Exception):
+    '''Raised if a host name does not resolve to an ip address.
+
+    Dugong raises this exception if a resolution attempt results in a
+    `socket.gaierror` or `socket.herror` exception with errno
+    :const:`!socket.EAI_AGAIN` or :const:`!socket.EAI_NONAME`, but at least one of
+    the hostnames in `DNS_TEST_HOSTNAMES` can be resolved.
+    '''
+
+    def __init__(self, hostname):
+        self.name = hostname
+
+    def __str__(self):
+        return 'Host %s does not have any ip addresses' % self.name
+
+
+class DNSUnavailable(Exception):
+    '''Raised if the DNS server cannot be reached.
+
+    Dugong raises this exception if a resolution attempt results in a
+    `socket.gaierror` or `socket.herror` exception with errno
+    :const:`!socket.EAI_AGAIN` or :const:`!socket.EAI_NONAME`, and none of the
+    hostnames in `DNS_TEST_HOSTNAMES` can be resolved either.
+    '''
+
+    def __init__(self, hostname):
+        self.name = hostname
+
+    def __str__(self):
+        return 'Unable to resolve %s, DNS server unavailable.' % self.name
 
 class StateError(_GeneralError):
     '''
@@ -401,11 +431,12 @@ class HTTPConnection:
 
         if self.proxy:
             log.debug('connecting to %s', self.proxy)
-            self._sock = socket.create_connection(self.proxy)
-            eval_coroutine(self._co_tunnel(), self.timeout)
+            self._sock = create_socket(self.proxy)
+            if self.ssl_context:
+                eval_coroutine(self._co_tunnel(), self.timeout)
         else:
             log.debug('connecting to %s', (self.hostname, self.port))
-            self._sock = socket.create_connection((self.hostname, self.port))
+            self._sock = create_socket((self.hostname, self.port))
 
         if self.ssl_context:
             log.debug('establishing ssl layer')
@@ -552,7 +583,11 @@ class HTTPConnection:
         headers['Accept-Encoding'] = 'identity'
         if 'Connection' not in headers:
             headers['Connection'] = 'keep-alive'
-        request = [ '{} {} HTTP/1.1'.format(method, path).encode('latin1') ]
+        if self.proxy and not self.ssl_context:
+            gpath = "http://{}{}".format(headers['Host'], path)
+        else:
+            gpath = path
+        request = [ '{} {} HTTP/1.1'.format(method, gpath).encode('latin1') ]
         for key, val in headers.items():
             request.append('{}: {}'.format(key, val).encode('latin1'))
         request.append(b'')
@@ -593,7 +628,7 @@ class HTTPConnection:
                 if self._sock is None:
                     raise ConnectionClosed('connection has been closed locally')
                 continue
-            except BrokenPipeError:
+            except (BrokenPipeError, ConnectionResetError):
                 raise ConnectionClosed('found closed when trying to write')
             except OSError as exc:
                 if exc.errno == errno.EINVAL:
@@ -975,9 +1010,10 @@ class HTTPConnection:
         len_ = min(len_, self._in_remaining)
         log.debug('updated len_=%d', len_)
 
-        # Buffer might be empty, but have no capacity. This is handled
-        # in _try_fill_buffer(), but we have to take this into account
-        # in the while condition already.
+        # If buffer is empty, reset so that we start filling from
+        # beginning. This check is already done by _try_fill_buffer(), but we
+        # have to do it here or we never enter the while loop if the buffer
+        # is empty but has no capacity (rbuf.b == rbuf.e == len(rbuf.d))
         if rbuf.b == rbuf.e:
             rbuf.b = 0
             rbuf.e = 0
@@ -1197,20 +1233,24 @@ class HTTPConnection:
         log.debug('start')
         rbuf = self._rbuf
 
-        # If no capacity or empty, reset
-        # (this clause cover *both* cases)
-        if rbuf.d == rbuf.e:
+        # If buffer is empty, reset so that we start filling from beginning
+        if rbuf.b == rbuf.e:
+            rbuf.b = 0
             rbuf.e = 0
-            rbuf.d = 0
+
+        # If no capacity, return
+        if rbuf.e == len(rbuf.d):
+            return 0
 
         try:
             len_ = self._sock.recv_into(memoryview(rbuf.d)[rbuf.e:])
         except (socket.timeout, ssl.SSLWantReadError, BlockingIOError):
             log.debug('done (nothing ready)')
             return None
+        except (ConnectionResetError, BrokenPipeError):
+            len_ = 0
 
         if not len_:
-            assert rbuf.e < len(rbuf.d)
             raise ConnectionClosed('connection closed unexpectedly')
 
         rbuf.e += len_
@@ -1221,6 +1261,8 @@ class HTTPConnection:
         '''Make sure that there are at least *len_* bytes in buffer'''
 
         rbuf = self._rbuf
+        if len_ > len(rbuf.d):
+            raise ValueError('Requested more bytes than buffer has capacity')
         sock_fd = self._sock.fileno()
         while len(rbuf) < len_:
             if len(rbuf.d) - rbuf.b < len_:
@@ -1356,35 +1398,68 @@ def eval_coroutine(crt, timeout=None):
     except StopIteration as exc:
         return exc.value
 
+def create_socket(address):
+    '''Create socket connected to *address**
+
+    Use `socket.create_connection` to create a connected socket and return
+    it. If a DNS related exception is raised, capture it and attempt to
+    determine if the dns server is not reachable, or if the host name could
+    not be resolved. Then raise either `NoSuchAddress` or
+    `NameResolutionError` as appropriate.
+
+    To distinguish between an unresolvable hostname and a problem with the
+    DNS server, attempt to resolve the addresses in `DNS_TEST_HOSTNAMES`. If
+    at least one test hostname can be resolved, assume that the DNS server
+    is available and that *address* can not be resolved.
+    '''
+
+    try:
+        return socket.create_connection(address)
+
+    # The exception unfortunately does not help us to distinguish between
+    # permanent and temporary problems. See:
+    # https://stackoverflow.com/questions/24855168/
+    # https://stackoverflow.com/questions/24855669/
+    except (socket.gaierror, socket.herror) as exc:
+        if exc.errno not in (socket.EAI_AGAIN, socket.EAI_NONAME):
+            raise
+
+    # Try to resolve test hosts
+    for (hostname, port) in DNS_TEST_HOSTNAMES:
+        try:
+            socket.getaddrinfo(hostname, port)
+        except (socket.gaierror, socket.herror) as exc:
+            if exc.errno not in (socket.EAI_AGAIN, socket.EAI_NONAME):
+                raise
+            # Not reachable, try next one
+        else:
+            # Reachable, now try to resolve original address
+            # again (maybe dns was only down briefly)
+            break
+    else:
+        # No host was reachable
+        raise DNSUnavailable(address[0])
+
+    # Try to connect to original host again
+    try:
+        return socket.create_connection(address)
+    except (socket.gaierror, socket.herror) as exc:
+        if exc.errno not in (socket.EAI_AGAIN, socket.EAI_NONAME):
+            raise
+        raise HostnameNotResolvable(address[0])
+
+
 def is_temp_network_error(exc):
     '''Return true if *exc* represents a potentially temporary network problem
 
-    For problems with name resolution, the exception generally does not contain
-    enough information to distinguish between an unresolvable hostname and a
-    problem with the DNS server. In this case, this function attempts to resolve
-    the addresses in `DNS_TEST_HOSTNAMES`. A name resolution problem is
-    considered permanent if at least one test hostname can be resolved, and
-    temporary if none of the test hostnames can be resolved.
+    DNS resolution errors (`socket.gaierror` or `socket.herror`) are considered
+    permanent, because `HTTPConnection` employs a heuristic to convert these
+    exceptions to `HostnameNotResolvable` or `DNSUnavailable` instead.
     '''
 
     if isinstance(exc, (socket.timeout, ConnectionError, TimeoutError, InterruptedError,
                         ConnectionClosed, ssl.SSLZeroReturnError, ssl.SSLEOFError,
-                        ssl.SSLSyscallError, ConnectionTimedOut)):
-        return True
-
-    # The exception also unfortunately does not help us to distinguish between
-    # permanent and temporary problems. See:
-    # https://stackoverflow.com/questions/24855168/
-    # https://stackoverflow.com/questions/24855669/
-    elif (isinstance(exc, (socket.gaierror, socket.herror))
-          and exc.errno in (socket.EAI_AGAIN, socket.EAI_NONAME)):
-        for (hostname, port) in DNS_TEST_HOSTNAMES:
-            try:
-                socket.getaddrinfo(hostname, port)
-            except (socket.gaierror, socket.herror):
-                pass
-            else:
-                return False
+                        ssl.SSLSyscallError, ConnectionTimedOut, DNSUnavailable)):
         return True
 
     return False
